/*
 * Copyright 2000-2015 JetBrains s.r.o.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.jetbrains.plugins.groovy.lang.psi.util;

import com.intellij.util.containers.ContainerUtil;
import org.jetbrains.annotations.NonNls;

import java.util.Collections;
import java.util.Set;

/**
 * @author Maxim.Medvedev
 */
public final class GroovyCommonClassNames {

  @NonNls public static final String GROOVY_OBJECT_SUPPORT = "groovy.lang.GroovyObjectSupport";
  @NonNls public static final String GROOVY_OBJECT = "groovy.lang.GroovyObject";
  @NonNls public static final String GROOVY_LANG_CLOSURE = "groovy.lang.Closure";
  @NonNls public static final String DEFAULT_BASE_CLASS_NAME = "groovy.lang.GroovyObject";
  @NonNls public static final String GROOVY_LANG_GSTRING = "groovy.lang.GString";
  @NonNls public static final String DEFAULT_GROOVY_METHODS = "org.codehaus.groovy.runtime.DefaultGroovyMethods";
  @NonNls public static final String GROOVY_LANG_SCRIPT = "groovy.lang.Script";
  @NonNls public static final String GROOVY_LANG_INT_RANGE = "groovy.lang.IntRange";
  @NonNls public static final String GROOVY_LANG_OBJECT_RANGE = "groovy.lang.ObjectRange";
  @NonNls public static final String GROOVY_LANG_DELEGATE = "groovy.lang.Delegate";
  @NonNls public static final String GROOVY_UTIL_CONFIG_OBJECT = "groovy.util.ConfigObject";
  @NonNls public static final String JAVA_UTIL_REGEX_PATTERN = "java.util.regex.Pattern";
  @NonNls public static final String JAVA_MATH_BIG_DECIMAL = "java.math.BigDecimal";
  @NonNls public static final String JAVA_MATH_BIG_INTEGER = "java.math.BigInteger";
  @NonNls public static final String ORG_CODEHAUS_GROOVY_RUNTIME_METHOD_CLOSURE = "org.codehaus.groovy.runtime.MethodClosure";
  @NonNls public static final String JAVA_UTIL_REGEX_MATCHER = "java.util.regex.Matcher";
  @NonNls public static final String GROOVY_TRANSFORM_FIELD = "groovy.transform.Field";
  @NonNls public static final String GROOVY_TRANSFORM_TUPLE_CONSTRUCTOR = "groovy.transform.TupleConstructor";
  @NonNls public static final String GROOVY_TRANSFORM_IMMUTABLE = "groovy.transform.Immutable";
  @NonNls public static final String GROOVY_TRANSFORM_CANONICAL = "groovy.transform.Canonical";
  @NonNls public static final String GROOVY_LANG_REFERENCE = "groovy.lang.Reference";
  @NonNls public static final String JAVA_UTIL_LINKED_HASH_MAP = "java.util.LinkedHashMap";
  @NonNls public static final String GROOVY_TRANSFORM_AUTO_EXTERNALIZE = "groovy.transform.AutoExternalize";
  @NonNls public static final String GROOVY_TRANSFORM_AUTO_CLONE = "groovy.transform.AutoClone";
  @NonNls public static final String GROOVY_LANG_CATEGORY = "groovy.lang.Category";
  @NonNls public static final String GROOVY_LANG_USE = "groovy.lang.Use";
  @NonNls public static final String GROOVY_LANG_MIXIN = "groovy.lang.Mixin";
  @NonNls public static final String GROOVY_UTIL_TEST_CASE = "groovy.util.GroovyTestCase";
  @NonNls public static final String GROOVY_LANG_SINGLETON = "groovy.lang.Singleton";
  @NonNls public static final String GROOVY_TRANSFORM_COMPILE_STATIC = "groovy.transform.CompileStatic";
  @NonNls public static final String GROOVY_TRANSFORM_TYPE_CHECKED = "groovy.transform.TypeChecked";
  @NonNls public static final String GROOVY_TRANSFORM_TYPE_CHECKING_MODE = "groovy.transform.TypeCheckingMode";
  @NonNls public static final String JAVA_LANG_CHAR_SEQUENCE = "java.lang.CharSequence";
  @NonNls public static final String GROOVY_TRANSFORM_INHERIT_CONSTRUCTORS = "groovy.transform.InheritConstructors";
  @NonNls public static final String GROOVY_LANG_IMMUTABLE = "groovy.lang.Immutable";
  @NonNls public static final String GROOVY_LANG_META_CLASS = "groovy.lang.MetaClass";
  @NonNls public static final String GROOVY_LANG_GROOVY_CALLABLE = "groovy.lang.GroovyCallable";
  @NonNls public static final String GROOVY_TRANSFORM_ANNOTATION_COLLECTOR = "groovy.transform.AnnotationCollector";
  @NonNls public static final String GROOVY_LANG_NEWIFY = "groovy.lang.Newify";
  @NonNls public static final String GROOVY_LANG_DELEGATES_TO = "groovy.lang.DelegatesTo";
  @NonNls public static final String GROOVY_LANG_DELEGATES_TO_TARGET = "groovy.lang.DelegatesTo.Target";
  @NonNls public static final String GROOVY_TRANSFORM_COMPILE_DYNAMIC = "groovy.transform.CompileDynamic";
  @NonNls public static final String GROOVY_TRANSFORM_STC_CLOSURE_PARAMS = "groovy.transform.stc.ClosureParams";
  @NonNls public static final String GROOVY_TRANSFORM_BASE_SCRIPT = "groovy.transform.BaseScript";
  @NonNls public static final String IMPLEMENTED_FQN = "org.codehaus.groovy.transform.trait.Traits.Implemented";

  public static final Set<String> GROOVY_EXTENSION_CLASSES = Collections.unmodifiableSet(ContainerUtil.newLinkedHashSet(
    "org.codehaus.groovy.runtime.DateGroovyMethods",
    "org.codehaus.groovy.runtime.DefaultGroovyMethods",
    "org.codehaus.groovy.runtime.DefaultGroovyStaticMethods",
    "org.codehaus.groovy.runtime.EncodingGroovyMethods",
    "org.codehaus.groovy.runtime.IOGroovyMethods",
    "org.codehaus.groovy.runtime.ProcessGroovyMethods",
    "org.codehaus.groovy.runtime.ResourceGroovyMethods",
    "org.codehaus.groovy.runtime.SocketGroovyMethods",
    "org.codehaus.groovy.runtime.SqlGroovyMethods",
    "org.codehaus.groovy.runtime.StringGroovyMethods",
    "org.codehaus.groovy.runtime.SwingGroovyMethods",
    "org.codehaus.groovy.runtime.XmlGroovyMethods",
    "org.codehaus.groovy.vmplugin.v5.PluginDefaultGroovyMethods"
  ));

  private GroovyCommonClassNames() {
  }
}
